' T13-Sensorkeyer mit Sinusausgang
'
' Pinbelegung ATtiny13:
' PB0 PWM-Ausgang,  PB1 optionaler Schaltausgang fuer TRX-Tastung,  PB2 frei (Sensor-ADC-Referenz),
' PB3 Punktsensor, PB4 Strichsensor
'
' Speed-Poti liegt an Adc0 / Reset. Werte unter halber Betriebsspannung loesen also Reset aus. Ist kein Bug,
' sondern Feature; dient zum Neukalibrieren bzw. zum Einschalten des Keyers.

' Ausgeschaltet wird ueber zweckentfremdeten Watchdog nach etwa 30 min Untaetigkeit

' Stromverbrauch mit 5V: 4mA / 6mA
' Stromverbrauch mit 2,5V: 1,7mA / 2,5 mA  
'
' Lizenz: WTFPL - http://www.wtfpl.net/about/
' verwendete Programmierungebung: BASCOM 2.0.7.9 Freeware-Version


$regfile = "ATtiny13.dat"
$crystal = 9600000                                          ' Taktfrequenz: 9,6 MHz
$hwstack = 8                                                ' Platz fuer nur 4 Unterprogramme reserviert, spart SRAM
$swstack = 0                                                ' wird nicht gebraucht
$framesize = 0                                              ' dito

Clkpr = 128                                                 ' Taktumschaltung auf 9,6 MHz
Clkpr = 0
Clkpr = 0

Dim Dit As Word                                             ' Laenge Punkt
Dim Dah As Word                                             ' Laenge Strich
Dim Sndlen As Word                                          ' Zaehlvariable fuer Zeichen-/Pausendauer
Dim P As Byte                                               ' "Punktspeicher"
Dim N As Byte                                               ' Schleifenvariable fuer Sinus-Erzeugung
Dim Anadah As Word                                          ' ADC-Wert fuer Erkennung Dah-Sensor
Dim Anadit As Word                                          ' ADC-Wert fuer Erkennung Dit-Sensor
Dim Sa As Byte                                              ' Amplitudenwert fuer PWM
Dim Refana As Word                                          ' speichert nach Reset den Sensor-Leerlauf-Kalibrierwert 
Dim Pause As Byte                                           ' wenn 1, wird in Sinus-Unterprogramm PWM = 0 gesetzt
Dim Offtimer As Byte                                        ' Zaehler fuer Abschaltautomatik

Config Timer0 = Pwm , Prescale = 1 , Compare A Pwm = Clear_up
'                                                           ' "Fast PWM" (37,5 kHz) per Registerbits konfigurieren:
Tccr0a.0 = 1                                                ' wgm00
Tccr0a.1 = 1                                                ' wgm01
Tccr0b.3 = 0                                                ' wgm02
Start Timer0

Acsr.acd = 1                                                ' Analog-Komparator ausschalten, spart etwas Strom

Config Adc = Single , Prescaler = Auto                      ' AD-Converter konfigurieren und starten
Start Adc

Ddrb = &B0000_0111                                          ' DDRB.0 ist Tonausgang; DDRB.2 ist der Pin der Sensor-ADC-Referenz
Portb = &B1110_0000                                         ' koennte man weglassen; ist eh Auslieferungszustand
Didr0 = &B00111111                                          ' Digitaleingaenge abschalten; spart Strom


On Watchdog Onwdt                                           ' Interrupt-Routine fuer Watchdog-interrupt 

' Watchdog als Interrupt-Quelle konfigurieren -      WDCR:  7=WDTIF  6=WDIE  5=WDP3  4=WDCE  3=WDE  2=WDP2  1=WDP1  0=WDP0
Wdtcr = &B0001_1111                                         ' Schreiben auf WDTCR Vorbereiten
Wdtcr = &B0110_0001                                         ' Interrupt statt Reset, Watchdog- Timer auf 8 sec setzen
Sreg.7 = 1                                                  ' Interrupts global freigeben


Waitms 200                                                  ' Vorlaufzeit fuer den ADC
Gosub Ditsensor                                             ' einmal ADC-Leerlaufwert am Punktsensor messen (Anpassung an Sensor)
Refana = Anadit + 96                                        ' Referenz 96 hoeher setzen


Do
    Gosub Adclesen
    Gosub Ditsensor
    If Anadit > Refana Or P = 1 Then                        ' Punktsensor gedrueckt (PB3) oder Punkt gespeichert
        P = 0                                               ' Punktspeicher loeschen
        Pause = 0
        Sndlen = Dit                                        ' Zeichenelement mit Laenge "Dit" senden
        Portb.1 = 1                                         ' Schaltausgang ein
        Gosub Sinus
        Portb.1 = 0                                         ' Schaltausgang aus 
        Pause = 1
        Sndlen = Dit                                        ' Pause nach dem Punkt
        Gosub Sinus
        Offtimer = 0                                        ' Watchdog-Zaehler bei Aktivitaet zuruecksetzen
    End If

    Gosub Dahsensor
    If Anadah > Refana Then                                 ' Strichsensor gedrueckt (PB4)
        Pause = 0
        Sndlen = Dah
        PortB.1 = 1
        Gosub Sinus
        PortB.1 = 0
        Gosub Ditsensor
        If Anadit > Refana Then                             ' registrieren, wenn am Strichende bereits Punkt gedrueckt ist
            P = 1                                           ' Punkt speichern
        End If
        Pause = 1
        Sndlen = Dit                                        ' Pause nach Strich
        Gosub Sinus
    End If
Loop


Sinus:
    Do
        Restore Sinusdata                                   ' Zeiger auf Anfang der Data-Tabelle setzen
        For N = 1 To 64
            If Pause = 0 Then
                Read Sa                                     ' Tabellenwert lesen
                Pwm0a = Sa                                  ' der Wert in PWM0A bestimmt das Puls- / Pausenverhältnis
            Else
                Pwm0a = 0                                   ' statt Nachschlagen in Tabelle konstant 0
                Waitus 3                                    ' Ausgleich dafuer, dass das Nachschlagen etwas laenger dauert
            End If
            Waitus 16                                       ' Frequenz des Tones senken;  16µs -> 600 Hz
        Next N
        Decr Sndlen                                         ' Software-Timer fuer Punkt- / Strichlaenge
    Loop Until Sndlen = 0                                   ' Zeichen fertig
Return


Ditsensor:
    Admux = 1                                               ' ADC1 (Referenz) an ADC-Eingang legen ; der muss auf Low liegen (bereits zu Anfang fest konfiguriert)
'    Portb.2 = 0                                            ' ADC1 auf 0 ziehen
'    DDRB.2 = 1                                             ' ADC1 an Ausgang legen --> Referenz entladen
    Portb.3 = 1                                             ' Pullup an Punktsensor ein
    Ddrb.3 = 1                                              ' Punktsensor nun Ausgang  -->  Sensorpin laden
    Waitus 32
    Ddrb.3 = 0                                              ' Punktsensor wieder Eingang
    Portb.3 = 0                                             ' Pullup an Punktsensor aus --> Sensorpin floatet
    Anadit = Getadc(3)                                      ' Punktsensor auslesen
Return

Dahsensor:
    Admux = 1                                               ' ADC1 (Referenz) an ADC-Eingang legen
    Portb.4 = 1                                             ' Pullup an Strichsensor ein
    Ddrb.4 = 1                                              ' Strichsensor ist nun Ausgang
    Waitus 32
    Ddrb.4 = 0                                              ' Strichsensor wieder Eingang
    Portb.4 = 0                                             ' Pullup an Strichsensor aus
    Anadah = Getadc(2)                                      ' Strichsensor auslesen
Return


Adclesen:
' Dit = 20: ~40 WpM   Dit = 24: 25 WpM   Dit = 64  11 WpM
' Mit 1280 und Rechts-Shift um 4: 17 WPM bis 50 WpM
    Dit = Getadc(0)
    Dit = 1280 - Dit                                        ' 1280 = 1024 + 256 
    Shift Dit , Right , 4                                   ' entspricht Teilung durch 16, spart Flash-Speicher
    Dah = Dit + Dit                                         ' statt Multiplikation; spart Flash (78 bytes!)
    Dah = Dah + Dit
Return


Onwdt:
    Incr Offtimer                                           ' alle 8 sec wird Offtimer incrementiert
    If Offtimer > 250 Then                                  ' nach 250 * 8 sec. erfolgt Powerdown
        Stop Adc
        DDRB = 0
'        Powerdown
        Mcucr = &B0011_0000                                  ' 4=1, 3=0: Powerdown   5=1 :Sleep enable
        !Sleep
    End If
Return


End

Sinusdata:

Data 0 , 0 , 2 , 5 , 9 , 14 , 21 , 28 , 36 , 46 , 56 , 66 , 78 , 89 , 101 , 114,
Data 127 , 139 , 151 , 163 , 175 , 186 , 197 , 207 , 216 , 225 , 232 , 238 , 244 , 248 , 251 , 253,
Data 253 , 253 , 251 , 248 , 244 , 239 , 232 , 225 , 216 , 207 , 197 , 187 , 175 , 164 , 151 , 139,
Data 127 , 114 , 102 , 90 , 78 , 67 , 56 , 46 , 37 , 28 , 21 , 15 , 9 , 5 , 2 , 0,
